/*
	level1.c -- Level 1 synchronous transmission routines

  Poor Man's Packet (PMP)
  Copyright (c) 1991 by Andrew C. Payne    All Rights Reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation without fee for NON-COMMERCIAL AMATEUR RADIO USE ONLY is hereby
  granted, provided that the above copyright notice appear in all copies.
  The author makes no representations about the suitability of this software
  for any purpose.  It is provided "as is" without express or implied warranty.

	07/24/89
	 Andrew C. Payne
*/

/* ----- Includes ----- */
#include <stdio.h>
#include <alloc.h>
#include <dos.h>
#include "pmp.h"
#include "crc.h"
#include "serial.h"


#define MAXTXQUEUE      50	/* max size of TXQueue */

/* ----- Local Variables ----- */

static struct ax25_level1 *TXQueue[MAXTXQUEUE];	/* Transmit queue */
static int	TXQSize;			/* number of items in queue */
       int	TXQBytes;			/* total bytes in queue */


/* ----- CRC Generation and Checking ----- */
/* CRCCheck(packet)
	Given a packet, generates a CRC, checks with existing CRC, and sets
	valid CRC.
*/
int CRCCheck(struct ax25_level1 *packet)
{
	word	crc;
	int	len;		/* packet length, data bytes */
	int	i;
	byte	*p,t;
	word	q;

	p = packet->data;	/* pointer to data */
	len = packet->len-2;	/* length less CRC bytes */

/* compute CRC */
	crc = 0xffff;		/* initialize */
	for(i=0; i<len; i++) {
		t = *p++ ^ (crc & 0xff);
		crc >>= 8;
		crc ^= crc_table[t];
	}
	q = ~crc;

	crc = *((word *)p);	/* fetch existing */
	*((word *)p) = q;	/* store computed one */

	return crc == q;	/* return OK or not OK */
}


/* ----- Transmitter Control ----- */

/* TXKey(state)
	Sets the transmitter state to the state given (TRUE = 1 = keyup).
	Updates the status line on the screen.
*/
static void TXKey(int state)
{
	// DB: Rather meaningless for KISS...
	ShowTXRX(state,FALSE);
}


/* sendpacketdata(p)
	Given a data packet, sends data packet.
*/
static void sendpacketdata(struct ax25_level1 *packet)
{
	int	len;			/* bytes left to send */
	byte	*p;			/* pointer to data */


/* initialize pointers and counters */
	len = packet->len;
	p = packet->data;

/* send the packet in KISS mode to the serial port. */
// Skip the CRC bytes, which are added by the modem.
	len -= 2;   // Ignore CRC bytes at end.

	StartKissFrame();
	WriteKiss(0);  // TODO: support different ports?
	while(len--) {
		WriteKiss(*p++);
	}
	EndKissFrame();

}


/* ----- Transmit Queue Manipulation ----- */

/* TXQInit()
	Initializes the level 1 transmit queue.
*/
void TXQInit()
{
	TXQBytes = TXQSize = 0;		/* empty queue */
}

/* TXQAdd(packet)
	Adds a packet to the TXQueue.  Returns TRUE if successful, or
	FALSE if error.

	Computes and sets CRC for packet.

	'packet' is a pointer to an AX.25 level 1 packet record, that was
	alloced with the malloc() function.
*/
int TXQAdd(struct ax25_level1 *packet)
{
/* check for full queue */
	if(TXQSize >= MAXTXQUEUE)
		return FALSE;

/* compute and set CRC */
	(void) CRCCheck(packet);

/* add to TXQueue */
	TXQBytes += packet->len;
	TXQueue[TXQSize++] = packet;
	return TRUE;
}

/* TXQEmpty()
	Assumes a clear channel, keys transmitter and empties TX queue.

	Frees all TX queue entries.
*/
void TXQEmpty()
{
	int	i;

/* is there anything in the queue to send? */
	if(TXQSize) {

		TXKey(TRUE);		/* key up transmitter */

/* send all packets in the TX Queue */
		for(i=0; i<TXQSize; i++) {
			sendpacketdata(TXQueue[i]);
		}
		TXKey(FALSE);		/* drop transmitter key */


/* free all packets in the TX Queue */
		for(i=0; i<TXQSize; i++) {
#ifdef TRACE
			if(DebugMode)
				LogPacket(TXQueue[i],2); /* log the packet */
#endif
			free(TXQueue[i]);
		}
		TXCount += TXQSize;
		TXQBytes = TXQSize = 0;
	}
}

/* TXQBytesInQ()
	Returns the number of bytes currently in the Level 1 TX queue.
*/
int TXQBytesInQ(void)
{
	return TXQBytes;
}


/* ----- KISS Support ---------------- */

// These write the KISS data to the already-opened serial port.


extern void StartKissFrame(void)
{
	WriteSerial(FEND);

}


extern void WriteKiss(byte b)
{
	if (b == FEND) {
		WriteSerial(FESC);
		WriteSerial(TFEND);
	} else if (b == FESC) {
		WriteSerial(FESC);
		WriteSerial(TFESC);
	} else {
		WriteSerial(b);
	}
}


extern void EndKissFrame(void)
{
	WriteSerial(FEND);
}

