/*  SPY.C -- Show data
 *
 *  Copyright (C) 1991-1995 by Alef Null. All rights reserved.
 *  Author(s): J Vuori
 *  Modification(s):
 */

/*
  $Header: E:/HAM/EVM/SPY/RCS/spy.c 1.7 1997/02/06 10:14:48 dbraun Exp $

  $Log: spy.c $
  Revision 1.7  1997/02/06 10:14:48  dbraun
  Added support for DSPPORT environment variable, and cleaned up
  menu.

  Revision 1.6  1996/07/24 13:51:17  dbraun
  Made toggle commands instead on/off with U/L case letters,
  and added A/Z commands for different FFT averaging time constants

  Revision 1.5  1996/07/24 13:36:24  dbraun
  Expanded FFT display to occupy entire width of display.

  Revision 1.4  1996/07/18 14:40:14  dbraun
  No real change

  Revision 1.3  1996/07/18 14:37:25  dbraun
  No real change

  Revision 1.2  1996/07/18 14:36:22  dbraun
  No real change

  Revision 1.1  1996/07/18 14:35:19  dbraun
  Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <conio.h>
#include <ctype.h>
#include "dsply.h"
#include "fft.h"
#include "serial.h"

#define TRUE 1
#define FALSE 0


#define N	    512     // number of horisontal lines
#define LOGN	    9	    // log2(N)
#define STEP	    8

#define PORT	    1	    // default serial port number
#define BAUD	    19200   // baud rate

#define EXT   	    "DAT"   // file's extension


typedef enum { True = -1, False = 0 } Bool;

static struct {
    unsigned ctrlC:1;
    unsigned reso:1;
    unsigned cont:1;
    unsigned fft:1;
    unsigned fix:1;
    unsigned avg:1;
    unsigned avgCount;
} flags;

typedef struct {
    int      max, min;
    Bool     fComplex;
    COMPLEX *cp, data[N];
} SAMPLES;


/*
 * CTRL-C handler
 */
void cdecl CtrlCHandler(void) {
    flags.ctrlC = 1;
    signal(SIGINT, CtrlCHandler);
}


static char *AddExtension(char *FileName, char *Extension) {
    static char Name[13];
    char        *s;
    
    s = Name;
    while(*FileName && *FileName != '.')
	*s++ = *FileName++;
        
    if(*FileName)
	while((*s++ = *FileName++) != NULL);
    else {
	*s++ = '.';
    
	while((*s++ = *Extension++) != NULL);
    }
    
    return(Name);
}


static int GetDefaultSerialPort(void)
{
    char *portStr;
    int portNum;

    portStr = getenv("DSPPORT");
    if (portStr) {
        portNum = atoi(portStr);
        if (portNum > 0 && portNum <= 4)
            return(portNum);
	else {
	    fprintf(stderr, "Invalid DSPPORT value: %s\n", portStr);
	    exit(1);
        }
    }

    return(PORT); // Default value
}


/*
 * Wait (max 3s) for a character from the serial port
 *
 *  returns -1 if there is no character
 */
static int ReadSerialLine(void) {
    unsigned long start;
    int 	  c;

    #define BIOS_TIMER (volatile unsigned long far *)(0x46c)

    /* wait until timer changes its state */
    start = *BIOS_TIMER;
    while (start+54L != *BIOS_TIMER)	 // 54.9ms*54=3s
	if ((c = ReadSerial()) != -1)
	    return (c);

    return (-1);
}


/*
 * Read samples from the DSP CARD 4
 */
static Bool ReadSamples(Bool *fComplex, COMPLEX *frame) {
    int      c;
    COMPLEX *p;

    WriteSerial('S');

    if (ReadSerialLine() == 'P') {
	switch (ReadSerialLine()) {
	case -1:
	    return (False);
	    break;

	case 0:
	    *fComplex = False;
	    break;

	default:
	    *fComplex = True;
	    break;
	}

	if (*fComplex)
	    for (p = frame; p < &frame[256]; p++) {
		if ((p->x = ReadSerialLine()) == -1)
		    return (False);

		if ((c = ReadSerialLine()) == -1)
		    return (False);
		else
		    p->x |= (c<<8);

		if ((p->y = ReadSerialLine()) == -1)
		    return (False);

		if ((c = ReadSerialLine()) == -1)
		    return (False);
		else
		    p->y |= (c<<8);
	    }
	else
	    for (p = frame; p < &frame[512]; p++) {
		if ((p->x = ReadSerialLine()) == -1)
		    return (False);

		if ((c = ReadSerialLine()) == -1)
		    return (False);
		else
		    p->x |= (c<<8);

	    }

	return (True);
    }

    return (False);
}


static Bool nextl(FILE *fd, long *data) {
    int  c;
    char buff[80], *p;

    /* first flush leading space out */
    p = buff;
    while (c = getc(fd), !isdigit(c) && c != '-' && c != EOF);
    *p++ = c;
    if (c == EOF) return (False);

    /* then get the main body of the character */
    while (c = getc(fd), c != EOF && (isdigit(c) || c == '-'))
	*p++ = c;

    *p = '\0';

    *data = atol(buff) > 8388608L ? atol(buff)-16777216L : atol(buff);

    return (True);
}


/*
 * Reads one frame from the given file
 */
static Bool readFile(char *name, COMPLEX *frame) {
    FILE    *fp;
    long     d;
    COMPLEX *p;

    if ((fp = fopen(name, "r")) == NULL)
	return (False);

    p = frame;
    while (p < &frame[N])
	if (nextl(fp, &d))
	    p++->x = (int)d;
	else {
	    fclose(fp);
	    return (False);
	}

    fclose(fp);
    return (True);
}


/*
 * Writes one frame to the given file
 */
static Bool writeFile(char *name, COMPLEX *frame) {
    FILE    *fp;
    COMPLEX *p;

    if ((fp = fopen(name, "w")) == NULL)
	return (False);

    p = frame;
    while (p < &frame[N]) {
	fprintf(fp, "%d\n", p->x);
	p++;
    }

    fclose(fp);
    return (True);
}


/*
 * Search max and min values
 */
void searchMinMax(SAMPLES *s) {
    COMPLEX *p;

    s->min =  32767;
    s->max = -32767;

    for (p = s->data; p < &(s->data[N]); p++) {
	s->min = min(s->min, s->fComplex ? min(p->x, p->y) : p->x);
	s->max = max(s->max, s->fComplex ? max(p->x, p->y) : p->x);
    }
}

/* A temporary function to mess with the data */
void tweak(SAMPLES *s) {
    COMPLEX *p;
    static double lastx = 0.0;
    static double lasty = 0.0;

    for (p = s->data; p < &(s->data[N]); p++) {
	double tmp = p->x;
	p->x -= lastx;
	lastx = tmp;
    }

    if (s->fComplex)
	for (p = s->data; p < &(s->data[N]); p++) {
	    double tmp = p->y;
	    p->y -= lasty;
	    lasty = tmp;
	}
}



/* Used to hold smoothed data. */
SAMPLES avgSamples;

/* Smooth the (FFT) data. */
void average(SAMPLES *s, unsigned *avgCount) {
    COMPLEX *p;
    COMPLEX *q;
    double avgFactor;

    /* For the first sample, avgFactor is 1.0, then 0.5, 0.33, 0.25, etc.
       So each time, the previous avgCount sets of samples are averaged. */
    if (*avgCount < 10000)
        (*avgCount)++;
    avgFactor = 1.0/(double)(*avgCount);

    /* An avgFactor of 1.0 is no averaging at all, but
       we will still copy the data to avgSamples so that
       averaging can leter start with good initial values. */

    if (avgFactor > 1.0 || avgFactor <= 0.0)
        avgFactor = 1.0;

    for (p = s->data, q = avgSamples.data; p < &(s->data[N]); p++, q++) {
	q->x = p->x = (1.0-avgFactor)*q->x + avgFactor*p->x;
    }

    if (s->fComplex)
        for (p = s->data, q = avgSamples.data; p < &(s->data[N]); p++, q++) {
	    q->y = p->y = (1.0-avgFactor)*q->y + avgFactor*p->y;
	}
}



/*
 * Calculate spectrum of the given data block
 */
void spectrum(SAMPLES *s) {
    struct complex tmp[N];
    int            i;

    /* first transform windowed input samples */
    HanningWin(tmp, s->data, LOGN);
    fft(tmp, LOGN);

    /* then calculate the logarithm of the power */
    for (i = 0; i < N; i++) {
	double power = 
            tmp[Bit_Reverse(i, LOGN)].x*tmp[Bit_Reverse(i, LOGN)].x+
            tmp[Bit_Reverse(i, LOGN)].y*tmp[Bit_Reverse(i, LOGN)].y;

        /* Prevent underflow */
        if (power <= 0.0)
            power = 0.5;

	/* Each 10dB of power is 100 units. */
	s->data[i].x = (int)(100.0*log10(power));
    }

    /* N1OWU: For some reason, the above AlefNull code produces a
       spectrum that is a mirror-image of itself.  So, expand the
       lower half to fill the whole thing. */

    for (i = N/2-1; i >= 0; i--)
	s->data[2*i].x = s->data[2*i+1].x = s->data[i].x;
}


/*
 * Plot given block
 */
static void plotSamples(SAMPLES *s) {
    int     i, j = 0;
    COMPLEX val;

    PlotStart(s->min, s->max);
    for (i = 0; i < N && &(s->cp)[j] < &s->data[N]; i++) {
	if (flags.reso ? !(i % STEP) : 1) {
	    val = (s->cp)[j++];
	    if (flags.reso) {
		PlotCircle(0, i, val.x);
		if (s->fComplex)
		    PlotCircle(1, i, val.y);
	    }
	}

	PlotLine(0, i, val.x);
	if (s->fComplex)
	    PlotLine(1, i, val.y);
    }
}


int cdecl main(int argc, char *argv[]) {
    static char *logo[] = {
	"N1OWU's EVM version of Alef Null SPY (" __DATE__ ")",
	"[n]",
	"[A]",
	" ",
	"Q,ESC,.   exit from the program                        ",
	"R, W      read/write file                              ",
	"N, P      go forward/backward in zoomed-in mode        ",
	"S         read samples in non-continuous mode          ",
	"Modes (upper-case = ON, lower-case = OFF):             ",
	"  C: continuous  T: zoom-in     X: auto-scale          ",
	"  F: FFT         A: smooth FFT  Z: extra-smooth FFT    ",
	NULL
    };
    SAMPLES samples;
    char   *filename = "LOG", option;

    int     port = GetDefaultSerialPort();

    /* first parse arguments */
    argc--; argv++;
    while(argc > 0) {
	switch(**argv) {
	case '-':
	    switch(option = *(++(*argv))) {
	    /* port selection */
	    case 'p':
	    case 'P':
		port = atoi(++(*argv));
		break;

	    default:
		fprintf(stderr, "unknown option '%c', type 'spy ?' to get the allowed options\n", option);
		break;
	    }
	    break;

	case '?':
	    fprintf(stderr, "usage: spy [-p<portno>]\n");
	    fprintf(stderr, "\t-p<portno> uses the specified (1 (default) or 2) serial port\n");
	    return (0);

	default:
	    filename = *argv;
	    break;
	}

	argc--; argv++;
    }

    /* then open serial line */
    signal(SIGINT, CtrlCHandler);
    if (!OpenSerial(port, BAUD)) {
	fprintf(stderr, "can't open serial port COM%d\n", port);
	return (1);
    }

    /* Initialize run-time flags. */
    flags.cont = FALSE;
    flags.fix = TRUE;
    samples.min = -32000;
    samples.max =  32000;

    /* open graphic display */
    if (InitDisplay(N, logo)) {
	fprintf(stderr, "%s: no support for the needed graphics device (VGA)\n", __FILE__);
	CloseSerial();
	return (1);
    }

    /* try to show default file at the first */
    samples.cp	     = samples.data;
    samples.fComplex = False;
    if (readFile(AddExtension(filename, EXT), samples.data) == True) {
	searchMinMax(&samples);
	plotSamples(&samples);
    }

    /* finally obey user instructions */
    while (True) {
	/* terminate if user is in panic */
	if (flags.ctrlC)
	    goto end;

	/* if key pressed, obey the given command */
	if (kbhit())
	    switch (getch()) {
	    case 'R':
	    case 'r':
		if (readFile(AddExtension(filename, EXT), samples.data) == True) {
		    samples.fComplex = False;
		    if (flags.fft) {
			spectrum(&samples);
			if (flags.avg)
			    average(&samples, &flags.avgCount);
	                /* Each 10dB of power is 100 units. */
	                /* The maximum power is about +12dB */
			if (flags.fix)
			    samples.max = 1100;
			else
			    searchMinMax(&samples);
		        samples.min = samples.max - 500;
		    }
		    plotSamples(&samples);
	        } else
		    putchar('\x7');
	        break;

	    case 'W':
	    case 'w':
		if (writeFile(AddExtension(filename, EXT), samples.data) == False)
		    putchar('\x7');
		break;

	    case 'S':
	    case 's':
		if (ReadSamples(&samples.fComplex, samples.data) == True) {
		    if (flags.fft) {
			spectrum(&samples);
			if (flags.avg)
			    average(&samples, &flags.avgCount);
	                /* Each 10dB of power is 100 units. */
	                /* The maximum power is about +12dB */
		        if (flags.fix)
			    samples.max = 1100;
			else
			    searchMinMax(&samples);
		        samples.min = samples.max - 500;
		    } else if (!flags.fix)
			searchMinMax(&samples);

		    plotSamples(&samples);
		} else
		    putchar('\x7');
		break;

	    case 'N':
	    case 'n':
		samples.cp = min(&samples.data[N-(flags.reso ? STEP : N)-1], samples.cp+(flags.reso ? STEP : N));
		plotSamples(&samples);
		break;

	    case 'P':
	    case 'p':
		samples.cp = max(samples.data, samples.cp-(flags.reso ? STEP : N));
		plotSamples(&samples);
		break;

	    case 'T':
		flags.reso = TRUE;
		plotSamples(&samples);
		break;

	    case 't':
		flags.reso = FALSE;
		plotSamples(&samples);
		break;

	    case 'C':
		flags.cont = TRUE;
		break;

	    case 'c':
		flags.cont = FALSE;
		break;

	    case 'A':
		/* Normal averaging. */
		flags.avg = TRUE;
		flags.avgCount = 0;
		break;

	    case 'Z':
		/* Extra-slow averaging. */
		flags.avg = TRUE;
		flags.avgCount = 0;
		break;

	    case 'a':
	    case 'z':
		/* No averaging. */
		flags.avg = FALSE;
		flags.avgCount = 0;
		break;

	    case '+':
		if (flags.fix) {
			if (samples.max < 16000) {
				samples.max *= 2;
				samples.min = -samples.max;
			}
		}
		break;

	    case '-':
		if (flags.fix) {
			if (samples.max > 4) {
				samples.max /= 2;
				samples.min = -samples.max;
			}
		}
		break;

	    case 'X':
		flags.fix = TRUE;
		samples.min = -32000;
		samples.max =  32000;
		break;

	    case 'x':
		flags.fix = FALSE;
		break;

	    case 'f':
		flags.fft = FALSE;
		break;

	    case 'F':
		flags.fft = TRUE;
		spectrum(&samples);
	        if (flags.avg)
	            average(&samples, &flags.avgCount);
		/* Each 10dB of power is 100 units. */
		/* The maximum power is about +12dB */
		if (flags.fix)
		    samples.max = 1100;
		else
		    searchMinMax(&samples);
		samples.min = samples.max - 500;
		plotSamples(&samples);
		break;

	    case '.':
	    case 'Q':
	    case 'q':
	    case 0x1b:  /* ESC */
		goto end;
		break;

	    default:
		putchar('\x7');
		break;
	    }

	/* continuously sample if in continuous mode */
	if (flags.cont)
	    if (ReadSamples(&samples.fComplex, samples.data) == True) {
		/* tweak(&samples); */
		if (flags.fft) {
		    spectrum(&samples);
		    if (flags.avg)
			average(&samples, &flags.avgCount);
		    /* Each 10dB of power is 100 units. */
		    /* The maximum power is about +12dB */
		    if (flags.fix)
			samples.max = 1100;
		    else
			searchMinMax(&samples);
		    samples.min = samples.max - 500;
		} else if (!flags.fix)
		    searchMinMax(&samples);

		plotSamples(&samples);
	    } else {
		putchar('\x7');
		flags.cont = 0;
	    }
    }

end:
    ReleaseDisplay();
    CloseSerial();
    return (0);
}
